"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSourceFile = exports.defaultLanguages = void 0;
const prettyhtml = require("@starptech/prettyhtml");
const shared_1 = require("@volar/shared");
const vueSfc = require("@vue/compiler-sfc");
const reactivity_1 = require("@vue/reactivity");
const css = require("vscode-css-languageservice");
const node_1 = require("vscode-languageserver/node");
const dedupe = require("./utils/dedupe");
const string_1 = require("./utils/string");
const templateLsMainScript_1 = require("./virtuals/templateLsMainScript");
const templateLsScript_1 = require("./virtuals/templateLsScript");
const scriptRaw_1 = require("./virtuals/scriptRaw");
const styles_raw_1 = require("./virtuals/styles.raw");
const jsons_raw_1 = require("./virtuals/jsons.raw");
const templateLsTemplateScript_1 = require("./virtuals/templateLsTemplateScript");
const template_raw_1 = require("./virtuals/template.raw");
const untrack_1 = require("./utils/untrack");
exports.defaultLanguages = {
    template: 'html',
    script: 'js',
    style: 'css',
};
function createSourceFile(initialDocument, templateTsLs, scriptTsLs, context) {
    // sources
    const templateTsProjectVersion = reactivity_1.ref();
    const scriptTsProjectVersion = reactivity_1.ref();
    const vueDoc = reactivity_1.ref(initialDocument);
    const vueUri = vueDoc.value.uri;
    const descriptor = reactivity_1.reactive({
        template: null,
        script: null,
        scriptSetup: null,
        styles: [],
        customBlocks: [],
    });
    const lastUpdateChanged = {
        template: false,
        script: false,
        scriptSetup: false,
    };
    const templateScriptData = reactivity_1.reactive({
        projectVersion: undefined,
        context: [],
        components: [],
        props: [],
        setupReturns: [],
        htmlElements: [],
        componentItems: [],
        htmlElementItems: [],
    });
    const vueHtmlDocument = reactivity_1.computed(() => {
        return context.htmlLs.parseHTMLDocument(vueDoc.value);
    });
    const sfcErrors = reactivity_1.ref([]);
    // virtual scripts
    const _virtualStyles = styles_raw_1.useStylesRaw(context.ts, untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.styles), context);
    const virtualJsonBlocks = jsons_raw_1.useJsonsRaw(untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.customBlocks), context);
    const virtualTemplateRaw = template_raw_1.useTemplateRaw(untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.template), context);
    const templateData = reactivity_1.computed(() => {
        if (virtualTemplateRaw.pugDocument.value) {
            const pugDoc = virtualTemplateRaw.pugDocument.value;
            return {
                sourceLang: 'pug',
                html: pugDoc.htmlCode,
                htmlToTemplate: (htmlStart, htmlEnd) => {
                    const pugRange = pugDoc.sourceMap.getSourceRange2(htmlStart, htmlEnd);
                    if (pugRange) {
                        return pugRange.start;
                    }
                },
            };
        }
        if (descriptor.template) {
            return {
                sourceLang: 'html',
                html: descriptor.template.content,
                htmlToTemplate: (htmlStart, _) => htmlStart,
            };
        }
    });
    const templateLsTemplateScript = templateLsTemplateScript_1.useTemplateLsTemplateScript(untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.template), templateScriptData, _virtualStyles.textDocuments, _virtualStyles.sourceMaps, templateData, context);
    const cssLsStyles = {
        textDocuments: reactivity_1.computed(() => [templateLsTemplateScript.cssTextDocument.value, ..._virtualStyles.textDocuments.value].filter(shared_1.notEmpty)),
        sourceMaps: reactivity_1.computed(() => [templateLsTemplateScript.cssSourceMap.value, ..._virtualStyles.sourceMaps.value].filter(shared_1.notEmpty)),
    };
    const docLsScript = scriptRaw_1.useScriptRaw(untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.script));
    const docLsScriptSetup = scriptRaw_1.useScriptRaw(untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.scriptSetup));
    const templateLsScript = templateLsScript_1.useTemplateLsScript('template', context.ts, vueDoc, reactivity_1.computed(() => descriptor.script), reactivity_1.computed(() => descriptor.scriptSetup), reactivity_1.computed(() => { var _a; return (_a = templateData.value) === null || _a === void 0 ? void 0 : _a.html; }));
    const templateLsMainScript = templateLsMainScript_1.useTemplateLsMainScript(untrack_1.untrack(() => vueDoc.value), reactivity_1.computed(() => descriptor.script), reactivity_1.computed(() => descriptor.scriptSetup), reactivity_1.computed(() => descriptor.template));
    const scriptLsScript = templateLsScript_1.useTemplateLsScript('script', context.ts, vueDoc, reactivity_1.computed(() => descriptor.script), reactivity_1.computed(() => descriptor.scriptSetup), reactivity_1.computed(() => { var _a; return (_a = templateData.value) === null || _a === void 0 ? void 0 : _a.html; }));
    // map / set
    const templatetTsSourceMaps = reactivity_1.computed(() => {
        const result = [
            templateLsScript.sourceMap.value,
            templateLsScript.sourceMapForSuggestion.value,
            templateLsTemplateScript.sourceMap.value,
            templateLsMainScript.sourceMap.value,
            // scriptLsScript.sourceMap.value,
        ].filter(shared_1.notEmpty);
        return result;
    });
    const templateLsDocs = reactivity_1.computed(() => {
        const docs = new Map();
        docs.set(templateLsMainScript.textDocument.value.uri, templateLsMainScript.textDocument.value);
        docs.set(templateLsScript.textDocument.value.uri, templateLsScript.textDocument.value);
        if (templateLsScript.textDocumentForSuggestion.value)
            docs.set(templateLsScript.textDocumentForSuggestion.value.uri, templateLsScript.textDocumentForSuggestion.value);
        if (templateLsTemplateScript.textDocument.value)
            docs.set(templateLsTemplateScript.textDocument.value.uri, templateLsTemplateScript.textDocument.value);
        return docs;
    });
    const tsSourceMaps = reactivity_1.computed(() => [
        scriptLsScript.sourceMap.value,
        ...templatetTsSourceMaps.value,
    ]);
    update(initialDocument);
    // getters
    const getComponentCompletionData = useComponentCompletionData();
    const getDiagnostics = useDiagnostics();
    return {
        uri: vueUri,
        getTemplateTagNames: untrack_1.untrack(() => { var _a; return (_a = templateLsTemplateScript.templateCodeGens.value) === null || _a === void 0 ? void 0 : _a.tagNames; }),
        getTemplateAttrNames: untrack_1.untrack(() => { var _a; return (_a = templateLsTemplateScript.templateCodeGens.value) === null || _a === void 0 ? void 0 : _a.attrNames; }),
        getTextDocument: untrack_1.untrack(() => vueDoc.value),
        update: untrack_1.untrack(update),
        updateTemplateScript: untrack_1.untrack(updateTemplateScript),
        getComponentCompletionData: untrack_1.untrack(getComponentCompletionData),
        getDiagnostics: untrack_1.untrack(getDiagnostics),
        getScriptTsDocument: untrack_1.untrack(() => scriptLsScript.textDocument.value),
        getTsSourceMaps: untrack_1.untrack(() => tsSourceMaps.value),
        getCssSourceMaps: untrack_1.untrack(() => cssLsStyles.sourceMaps.value),
        getJsonSourceMaps: untrack_1.untrack(() => virtualJsonBlocks.sourceMaps.value),
        getHtmlSourceMaps: untrack_1.untrack(() => virtualTemplateRaw.htmlSourceMap.value ? [virtualTemplateRaw.htmlSourceMap.value] : []),
        getPugSourceMaps: untrack_1.untrack(() => virtualTemplateRaw.pugSourceMap.value ? [virtualTemplateRaw.pugSourceMap.value] : []),
        getTemplateScriptData: untrack_1.untrack(() => templateScriptData),
        getTeleports: untrack_1.untrack(() => [
            templateLsTemplateScript.teleportSourceMap.value,
            templateLsScript.teleportSourceMap.value,
        ].filter(shared_1.notEmpty)),
        getDescriptor: untrack_1.untrack(() => descriptor),
        getVueHtmlDocument: untrack_1.untrack(() => vueHtmlDocument.value),
        getTemplateLsDocs: untrack_1.untrack(() => templateLsDocs.value),
        getVirtualScript: untrack_1.untrack(() => ({
            document: templateLsScript.textDocument.value,
            sourceMap: templateLsScript.sourceMap.value,
        })),
        getScriptSetupData: untrack_1.untrack(() => templateLsScript.scriptSetupRanges.value),
        docLsScripts: untrack_1.untrack(() => ({
            documents: [docLsScript.textDocument.value, docLsScriptSetup.textDocument.value].filter(shared_1.notEmpty),
            sourceMaps: [docLsScript.sourceMap.value, docLsScriptSetup.sourceMap.value].filter(shared_1.notEmpty),
        })),
        getTemplateFormattingScript: untrack_1.untrack(() => ({
            document: templateLsTemplateScript.textDocumentForFormatting.value,
            sourceMap: templateLsTemplateScript.sourceMapForFormatting.value,
        })),
        shouldVerifyTsScript: untrack_1.untrack(shouldVerifyTsScript),
        refs: {
            cssSourceMaps: cssLsStyles.sourceMaps,
            htmlSourceMap: virtualTemplateRaw.htmlSourceMap,
            templateTsSourceMaps: templatetTsSourceMaps,
            templateTsDocuments: templateLsDocs,
            templateMainTsDocument: templateLsMainScript.textDocument,
            templaetTsTeleports: reactivity_1.computed(() => [
                templateLsTemplateScript.teleportSourceMap.value,
                templateLsScript.teleportSourceMap.value,
            ].filter(shared_1.notEmpty)),
            scriptTsDocument: scriptLsScript.textDocument,
            scriptTsTeleport: scriptLsScript.teleportSourceMap,
            scriptTsSourceMap: scriptLsScript.sourceMap,
        },
    };
    function update(newVueDocument) {
        var _a, _b, _c, _d;
        const parsedSfc = vueSfc.parse(newVueDocument.getText(), { sourceMap: false });
        const newDescriptor = parsedSfc.descriptor;
        const versionsBeforeUpdate = [
            (_a = templateLsScript.textDocument.value) === null || _a === void 0 ? void 0 : _a.version,
            (_b = templateLsTemplateScript.textDocument.value) === null || _b === void 0 ? void 0 : _b.version,
        ];
        updateSfcErrors();
        updateTemplate(newDescriptor);
        updateScript(newDescriptor);
        updateScriptSetup(newDescriptor);
        updateStyles(newDescriptor);
        updateCustomBlocks(newDescriptor);
        templateLsTemplateScript.update(); // TODO
        if (newVueDocument.getText() !== vueDoc.value.getText()) {
            vueDoc.value = newVueDocument;
        }
        const versionsAfterUpdate = [
            (_c = templateLsScript.textDocument.value) === null || _c === void 0 ? void 0 : _c.version,
            (_d = templateLsTemplateScript.textDocument.value) === null || _d === void 0 ? void 0 : _d.version,
        ];
        return {
            scriptUpdated: versionsBeforeUpdate[0] !== versionsAfterUpdate[0],
            templateScriptUpdated: versionsBeforeUpdate[1] !== versionsAfterUpdate[1],
        };
        function updateSfcErrors() {
            const errors = [];
            for (const error of parsedSfc.errors) {
                if ('code' in error && error.loc) {
                    const diag = node_1.Diagnostic.create(node_1.Range.create(error.loc.start.line - 1, error.loc.start.column - 1, error.loc.end.line - 1, error.loc.end.column - 1), error.message, node_1.DiagnosticSeverity.Error, error.code, 'vue');
                    errors.push(diag);
                }
            }
            sfcErrors.value = errors;
        }
        function updateTemplate(newDescriptor) {
            var _a, _b;
            const newData = newDescriptor.template ? {
                lang: (_a = newDescriptor.template.lang) !== null && _a !== void 0 ? _a : exports.defaultLanguages.template,
                content: newDescriptor.template.content,
                loc: {
                    start: newDescriptor.template.loc.start.offset,
                    end: newDescriptor.template.loc.end.offset,
                },
            } : null;
            lastUpdateChanged.template = ((_b = descriptor.template) === null || _b === void 0 ? void 0 : _b.content) !== (newData === null || newData === void 0 ? void 0 : newData.content);
            if (descriptor.template && newData) {
                descriptor.template.lang = newData.lang;
                descriptor.template.content = newData.content;
                descriptor.template.loc.start = newData.loc.start;
                descriptor.template.loc.end = newData.loc.end;
            }
            else {
                descriptor.template = newData;
            }
        }
        function updateScript(newDescriptor) {
            var _a, _b;
            const newData = newDescriptor.script ? {
                src: newDescriptor.script.src,
                lang: newDescriptor.script.src ? 'ts' : (_a = newDescriptor.script.lang) !== null && _a !== void 0 ? _a : exports.defaultLanguages.script,
                content: newDescriptor.script.src ? '' : newDescriptor.script.content,
                loc: {
                    start: newDescriptor.script.loc.start.offset,
                    end: newDescriptor.script.loc.end.offset,
                },
            } : null;
            lastUpdateChanged.script = ((_b = descriptor.script) === null || _b === void 0 ? void 0 : _b.content) !== (newData === null || newData === void 0 ? void 0 : newData.content);
            if (descriptor.script && newData) {
                descriptor.script.src = newData.src;
                descriptor.script.lang = newData.lang;
                descriptor.script.content = newData.content;
                descriptor.script.loc.start = newData.loc.start;
                descriptor.script.loc.end = newData.loc.end;
            }
            else {
                descriptor.script = newData;
            }
        }
        function updateScriptSetup(newDescriptor) {
            var _a, _b;
            const newData = newDescriptor.scriptSetup ? {
                lang: (_a = newDescriptor.scriptSetup.lang) !== null && _a !== void 0 ? _a : 'js',
                content: newDescriptor.scriptSetup.content,
                loc: {
                    start: newDescriptor.scriptSetup.loc.start.offset,
                    end: newDescriptor.scriptSetup.loc.end.offset,
                },
            } : null;
            lastUpdateChanged.scriptSetup = ((_b = descriptor.scriptSetup) === null || _b === void 0 ? void 0 : _b.content) !== (newData === null || newData === void 0 ? void 0 : newData.content);
            if (descriptor.scriptSetup && newData) {
                descriptor.scriptSetup.lang = newData.lang;
                descriptor.scriptSetup.content = newData.content;
                descriptor.scriptSetup.loc.start = newData.loc.start;
                descriptor.scriptSetup.loc.end = newData.loc.end;
            }
            else {
                descriptor.scriptSetup = newData;
            }
        }
        function updateStyles(newDescriptor) {
            var _a;
            for (let i = 0; i < newDescriptor.styles.length; i++) {
                const style = newDescriptor.styles[i];
                const newData = {
                    lang: (_a = style.lang) !== null && _a !== void 0 ? _a : exports.defaultLanguages.style,
                    content: style.content,
                    loc: {
                        start: style.loc.start.offset,
                        end: style.loc.end.offset,
                    },
                    module: !!style.module,
                    scoped: !!style.scoped,
                };
                if (descriptor.styles.length > i) {
                    descriptor.styles[i].lang = newData.lang;
                    descriptor.styles[i].content = newData.content;
                    descriptor.styles[i].loc.start = newData.loc.start;
                    descriptor.styles[i].loc.end = newData.loc.end;
                    descriptor.styles[i].module = newData.module;
                    descriptor.styles[i].scoped = newData.scoped;
                }
                else {
                    descriptor.styles.push(newData);
                }
            }
            while (descriptor.styles.length > newDescriptor.styles.length) {
                descriptor.styles.pop();
            }
        }
        function updateCustomBlocks(newDescriptor) {
            var _a;
            for (let i = 0; i < newDescriptor.customBlocks.length; i++) {
                const block = newDescriptor.customBlocks[i];
                const newData = {
                    type: block.type,
                    lang: (_a = block.lang) !== null && _a !== void 0 ? _a : '',
                    content: block.content,
                    loc: {
                        start: block.loc.start.offset,
                        end: block.loc.end.offset,
                    },
                };
                if (descriptor.customBlocks.length > i) {
                    descriptor.customBlocks[i].type = newData.type;
                    descriptor.customBlocks[i].lang = newData.lang;
                    descriptor.customBlocks[i].content = newData.content;
                    descriptor.customBlocks[i].loc.start = newData.loc.start;
                    descriptor.customBlocks[i].loc.end = newData.loc.end;
                }
                else {
                    descriptor.customBlocks.push(newData);
                }
            }
            while (descriptor.customBlocks.length > newDescriptor.customBlocks.length) {
                descriptor.customBlocks.pop();
            }
        }
    }
    function updateTemplateScript() {
        var _a, _b;
        const newVersion = (_b = (_a = templateTsLs.__internal__.host).getProjectVersion) === null || _b === void 0 ? void 0 : _b.call(_a);
        if (templateScriptData.projectVersion === newVersion) {
            return false;
        }
        templateScriptData.projectVersion = newVersion;
        const doc = templateLsMainScript.textDocument.value;
        const docText = doc.getText();
        const context = docText.indexOf(string_1.SearchTexts.Context) >= 0 ? templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(docText.indexOf(string_1.SearchTexts.Context))) : [];
        let components = docText.indexOf(string_1.SearchTexts.Components) >= 0 ? templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(docText.indexOf(string_1.SearchTexts.Components))) : [];
        const props = docText.indexOf(string_1.SearchTexts.Props) >= 0 ? templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(docText.indexOf(string_1.SearchTexts.Props))) : [];
        const setupReturns = docText.indexOf(string_1.SearchTexts.SetupReturns) >= 0 ? templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(docText.indexOf(string_1.SearchTexts.SetupReturns))) : [];
        const globalEls = docText.indexOf(string_1.SearchTexts.HtmlElements) >= 0 ? templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(doc.getText().indexOf(string_1.SearchTexts.HtmlElements))) : [];
        components = components.filter(entry => {
            const name = entry.data.name;
            return name.indexOf('$') === -1 && !name.startsWith('_');
        });
        const contextNames = context.map(entry => entry.data.name);
        const componentNames = components.map(entry => entry.data.name);
        const propNames = props.map(entry => entry.data.name);
        const setupReturnNames = setupReturns.map(entry => entry.data.name);
        const htmlElementNames = globalEls.map(entry => entry.data.name);
        if (shared_1.eqSet(new Set(contextNames), new Set(templateScriptData.context))
            && shared_1.eqSet(new Set(componentNames), new Set(templateScriptData.components))
            && shared_1.eqSet(new Set(propNames), new Set(templateScriptData.props))
            && shared_1.eqSet(new Set(setupReturnNames), new Set(templateScriptData.setupReturns))
            && shared_1.eqSet(new Set(htmlElementNames), new Set(templateScriptData.htmlElements))) {
            return false;
        }
        templateScriptData.context = contextNames;
        templateScriptData.components = componentNames;
        templateScriptData.props = propNames;
        templateScriptData.setupReturns = setupReturnNames;
        templateScriptData.htmlElements = htmlElementNames;
        templateScriptData.componentItems = components;
        templateScriptData.htmlElementItems = globalEls;
        templateLsTemplateScript.update(); // TODO
        return true;
    }
    function shouldVerifyTsScript(tsUri, mode) {
        var _a, _b;
        if (tsUri.toLowerCase() === ((_a = templateLsScript.textDocumentForSuggestion.value) === null || _a === void 0 ? void 0 : _a.uri.toLowerCase())) {
            if (mode === 3) {
                return 'all';
            }
            if (mode === 1) {
                const tsOptions = templateTsLs.__internal__.host.getCompilationSettings();
                const anyNoUnusedEnabled = tsOptions.noUnusedLocals || tsOptions.noUnusedParameters;
                return anyNoUnusedEnabled ? 'unused' : 'none';
            }
            return 'none';
        }
        if (tsUri.toLowerCase() === ((_b = templateLsScript.textDocument.value) === null || _b === void 0 ? void 0 : _b.uri.toLowerCase())) {
            if (mode === 3) {
                return !templateLsScript.textDocumentForSuggestion.value ? 'all' : 'none';
            }
            return 'all';
        }
        return 'all';
    }
    function useDiagnostics() {
        const tsOptions = templateTsLs.__internal__.host.getCompilationSettings();
        const anyNoUnusedEnabled = tsOptions.noUnusedLocals || tsOptions.noUnusedParameters;
        const nonTs = [
            [useStylesValidation(reactivity_1.computed(() => cssLsStyles.textDocuments.value)), 0, []],
            [useJsonsValidation(reactivity_1.computed(() => virtualJsonBlocks.textDocuments.value)), 0, []],
            [useTemplateValidation(), 0, []],
            [useScriptExistValidation(), 0, []],
        ];
        let templateTs = [
            [useTemplateScriptValidation(1), 0, []],
            [useTemplateScriptValidation(2), 0, []],
            [useTemplateScriptValidation(3), 0, []],
        ];
        let scriptTs = [
            [useScriptValidation(scriptLsScript.textDocument, 1), 0, []],
            [useScriptValidation(scriptLsScript.textDocument, 2), 0, []],
            [useScriptValidation(reactivity_1.computed(() => { var _a; return (_a = scriptLsScript.textDocumentForSuggestion.value) !== null && _a !== void 0 ? _a : scriptLsScript.textDocument.value; }), 3), 0, []],
            // [useScriptValidation(virtualScriptGen.textDocument, 4), 0, []], // TODO: support cancel because it's very slow
            [useScriptValidation(reactivity_1.computed(() => anyNoUnusedEnabled ? scriptLsScript.textDocumentForSuggestion.value : undefined), 1, true), 0, []],
        ];
        return async (response, isCancel) => {
            var _a, _b, _c, _d;
            templateTsProjectVersion.value = (_b = (_a = templateTsLs.__internal__.host).getProjectVersion) === null || _b === void 0 ? void 0 : _b.call(_a);
            scriptTsProjectVersion.value = (_d = (_c = scriptTsLs.__internal__.host).getProjectVersion) === null || _d === void 0 ? void 0 : _d.call(_c);
            // sort by cost
            templateTs = templateTs.sort((a, b) => a[1] - b[1]);
            scriptTs = scriptTs.sort((a, b) => a[1] - b[1]);
            let all = [...nonTs];
            let mainTsErrorStart = all.length - 1;
            let mainTsErrorEnd = -1;
            const isScriptChanged = lastUpdateChanged.script || lastUpdateChanged.scriptSetup;
            if (isScriptChanged) {
                all = all.concat(scriptTs);
                mainTsErrorEnd = all.length - 1;
                all = all.concat(templateTs);
            }
            else {
                all = all.concat(templateTs);
                mainTsErrorEnd = all.length - 1;
                all = all.concat(scriptTs);
            }
            let isDirty = false;
            for (let i = 0; i < all.length; i++) {
                if (await (isCancel === null || isCancel === void 0 ? void 0 : isCancel()))
                    return;
                const startTime = Date.now();
                const diag = all[i];
                if (!isDirty) {
                    isDirty = isErrorsDirty(diag[2], await diag[0].result.value);
                }
                diag[2] = await diag[0].result.value;
                diag[1] = Date.now() - startTime;
                const _newErrors = all
                    .slice(0, i + 1)
                    .map(async (diag) => await diag[0].result.value)
                    .flat();
                const newErrors = (await Promise.all(_newErrors))
                    .flat()
                    .concat(sfcErrors.value);
                const _oldErrors = all
                    .slice(i + 1)
                    .map(async (diag) => i >= mainTsErrorStart && !isScriptChanged ? await diag[0].cache.value : diag[2]);
                const oldErrors = (await Promise.all(_oldErrors))
                    .flat();
                ;
                const isLast = i === all.length - 1;
                if (await (isCancel === null || isCancel === void 0 ? void 0 : isCancel()))
                    return;
                if (isLast
                    || (isDirty && (i < mainTsErrorStart
                        || i === mainTsErrorEnd
                        || oldErrors.length === 0))) {
                    isDirty = false;
                    response(dedupe.withDiagnostics(newErrors.concat(oldErrors)));
                }
            }
            function isErrorsDirty(oldErrors, newErrors) {
                return !shared_1.eqSet(errorsToKeys(oldErrors), errorsToKeys(newErrors));
            }
            function errorsToKeys(errors) {
                return new Set(errors.map(error => error.source
                    + ':' + error.code
                    + ':' + error.message));
            }
        };
        function useTemplateValidation() {
            const htmlErrors = reactivity_1.computed(() => {
                if (virtualTemplateRaw.textDocument.value && virtualTemplateRaw.htmlDocument.value) {
                    return getVueCompileErrors(virtualTemplateRaw.textDocument.value);
                }
                return [];
            });
            const pugErrors = reactivity_1.computed(() => {
                var _a, _b;
                const result = [];
                if (virtualTemplateRaw.textDocument.value && virtualTemplateRaw.pugDocument.value) {
                    const pugDoc = virtualTemplateRaw.pugDocument.value;
                    const astError = pugDoc.error;
                    if (astError) {
                        result.push({
                            code: astError.code,
                            message: astError.msg,
                            range: {
                                start: { line: astError.line, character: astError.column },
                                end: { line: astError.line, character: astError.column },
                            },
                        });
                    }
                    else {
                        const htmlDoc = pugDoc.sourceMap.mappedDocument;
                        const vueCompileErrors = getVueCompileErrors(htmlDoc);
                        for (const vueCompileError of vueCompileErrors) {
                            let pugRange = pugDoc.sourceMap.getSourceRange(vueCompileError.range.start, vueCompileError.range.end);
                            if (!pugRange) {
                                const pugStart = (_a = pugDoc.sourceMap.getSourceRange(vueCompileError.range.start, vueCompileError.range.start)) === null || _a === void 0 ? void 0 : _a.start;
                                const pugEnd = (_b = pugDoc.sourceMap.getSourceRange(vueCompileError.range.end, vueCompileError.range.end)) === null || _b === void 0 ? void 0 : _b.end;
                                if (pugStart && pugEnd) {
                                    pugRange = {
                                        start: pugStart,
                                        end: pugEnd,
                                    };
                                    // trim empty space
                                    const pugText = pugDoc.sourceMap.sourceDocument.getText(pugRange);
                                    const trimLength = pugText.length - pugText.trimEnd().length;
                                    if (trimLength) {
                                        pugRange.end = pugDoc.sourceMap.sourceDocument.positionAt(pugDoc.sourceMap.sourceDocument.offsetAt(pugEnd)
                                            - trimLength);
                                    }
                                }
                            }
                            if (pugRange) {
                                vueCompileError.range = pugRange;
                                result.push(vueCompileError);
                            }
                            else {
                                let htmlText = htmlDoc.getText(vueCompileError.range);
                                let errorText = '';
                                try {
                                    errorText += '\n```html\n' + prettyhtml(htmlText).contents.trim() + '\n```'; // may thorw
                                }
                                catch (error) {
                                    errorText += '\n```html\n' + htmlText.trim() + '\n```'; // may thorw
                                    errorText += '\n```json\n' + JSON.stringify(error, null, 2) + '\n```';
                                }
                                vueCompileError.message += errorText;
                                vueCompileError.range = {
                                    start: virtualTemplateRaw.textDocument.value.positionAt(0),
                                    end: virtualTemplateRaw.textDocument.value.positionAt(virtualTemplateRaw.textDocument.value.getText().length),
                                };
                                result.push(vueCompileError);
                            }
                        }
                    }
                }
                return result;
            });
            const htmlErrors_cache = reactivity_1.ref([]);
            const pugErrors_cache = reactivity_1.ref([]);
            const result = reactivity_1.computed(() => {
                htmlErrors_cache.value = htmlErrors.value;
                pugErrors_cache.value = pugErrors.value;
                return cacheWithSourceMap.value;
            });
            const cacheWithSourceMap = reactivity_1.computed(() => {
                if (!virtualTemplateRaw.textDocument.value)
                    return [];
                return [
                    ...toSourceDiags(htmlErrors.value, virtualTemplateRaw.textDocument.value.uri, virtualTemplateRaw.htmlSourceMap.value ? [virtualTemplateRaw.htmlSourceMap.value] : []),
                    ...toSourceDiags(pugErrors.value, virtualTemplateRaw.textDocument.value.uri, virtualTemplateRaw.pugSourceMap.value ? [virtualTemplateRaw.pugSourceMap.value] : []),
                ];
            });
            return {
                result,
                cache: cacheWithSourceMap,
            };
            function getVueCompileErrors(doc) {
                const result = [];
                try {
                    const templateResult = vueSfc.compileTemplate({
                        source: doc.getText(),
                        filename: shared_1.uriToFsPath(vueUri),
                        id: shared_1.uriToFsPath(vueUri),
                        compilerOptions: {
                            onError: err => {
                                if (!err.loc)
                                    return;
                                const diagnostic = {
                                    range: {
                                        start: doc.positionAt(err.loc.start.offset),
                                        end: doc.positionAt(err.loc.end.offset),
                                    },
                                    severity: node_1.DiagnosticSeverity.Error,
                                    code: err.code,
                                    source: 'vue',
                                    message: err.message,
                                };
                                result.push(diagnostic);
                            },
                        }
                    });
                    for (const err of templateResult.errors) {
                        if (typeof err !== 'object' || !err.loc)
                            continue;
                        const diagnostic = {
                            range: {
                                start: doc.positionAt(err.loc.start.offset),
                                end: doc.positionAt(err.loc.end.offset),
                            },
                            severity: node_1.DiagnosticSeverity.Error,
                            source: 'vue',
                            code: err.code,
                            message: err.message,
                        };
                        result.push(diagnostic);
                    }
                }
                catch (err) {
                    const diagnostic = {
                        range: {
                            start: doc.positionAt(0),
                            end: doc.positionAt(doc.getText().length),
                        },
                        severity: node_1.DiagnosticSeverity.Error,
                        code: err.code,
                        source: 'vue',
                        message: err.message,
                    };
                    result.push(diagnostic);
                }
                return result;
            }
        }
        function useStylesValidation(documents) {
            const errors = reactivity_1.computed(() => {
                let result = new Map();
                for (const { textDocument, stylesheet } of documents.value) {
                    const cssLs = context.getCssLs(textDocument.languageId);
                    if (!cssLs || !stylesheet)
                        continue;
                    const errs = cssLs.doValidation(textDocument, stylesheet);
                    if (errs)
                        result.set(textDocument.uri, errs);
                }
                return result;
            });
            const errors_cache = reactivity_1.ref(new Map());
            const result = reactivity_1.computed(() => {
                errors_cache.value = errors.value;
                return cacheWithSourceMap.value;
            });
            const cacheWithSourceMap = reactivity_1.computed(() => {
                let result = [];
                for (const [uri, errs] of errors_cache.value) {
                    result = result.concat(toSourceDiags(errs, uri, cssLsStyles.sourceMaps.value));
                }
                return result;
            });
            return {
                result,
                cache: cacheWithSourceMap,
            };
        }
        function useJsonsValidation(documents) {
            const errors = reactivity_1.computed(async () => {
                var _a;
                let result = new Map();
                for (const { textDocument, jsonDocument } of documents.value) {
                    const errs = await context.jsonLs.doValidation(textDocument, jsonDocument, textDocument.languageId === 'json'
                        ? {
                            comments: 'error',
                            trailingCommas: 'error',
                        }
                        : {
                            comments: 'ignore',
                            trailingCommas: 'warning',
                        });
                    if (errs) {
                        for (const err of errs) {
                            err.source = (_a = err.source) !== null && _a !== void 0 ? _a : 'json';
                        }
                        result.set(textDocument.uri, errs);
                    }
                }
                return result;
            });
            const errors_cache = reactivity_1.ref();
            const result = reactivity_1.computed(() => {
                errors_cache.value = errors.value;
                return cacheWithSourceMap.value;
            });
            const cacheWithSourceMap = reactivity_1.computed(async () => {
                let result = [];
                if (errors_cache.value) {
                    for (const [uri, errs] of await errors_cache.value) {
                        result = result.concat(toSourceDiags(errs, uri, virtualJsonBlocks.sourceMaps.value));
                    }
                }
                return result;
            });
            return {
                result,
                cache: cacheWithSourceMap,
            };
        }
        function useScriptExistValidation() {
            const result = reactivity_1.computed(() => {
                const diags = [];
                if (!scriptTsLs.__internal__.getValidTextDocument(scriptLsScript.textDocument.value.uri)) {
                    for (const script of [descriptor.script, descriptor.scriptSetup]) {
                        if (!script || script.content === '')
                            continue;
                        const error = node_1.Diagnostic.create({
                            start: vueDoc.value.positionAt(script.loc.start),
                            end: vueDoc.value.positionAt(script.loc.end),
                        }, 'Virtual script not found, may missing lang="ts" or "allowJs": true.', node_1.DiagnosticSeverity.Information, undefined, 'volar');
                        error.tags = [node_1.DiagnosticTag.Unnecessary];
                        diags.push(error);
                    }
                }
                return diags;
            });
            return {
                result,
                cache: result,
            };
        }
        function useScriptValidation(document, mode, onlyUnusedCheck = false) {
            const errors = reactivity_1.computed(() => {
                if (mode === 1) { // watching
                    scriptTsProjectVersion.value;
                }
                const doc = document.value;
                if (!doc)
                    return [];
                if (mode === 1) {
                    return scriptTsLs.doValidation(doc.uri, { semantic: true });
                }
                else if (mode === 2) {
                    return scriptTsLs.doValidation(doc.uri, { syntactic: true });
                }
                else if (mode === 3) {
                    return scriptTsLs.doValidation(doc.uri, { suggestion: true });
                }
                else if (mode === 4) {
                    return scriptTsLs.doValidation(doc.uri, { declaration: true });
                }
                return [];
            });
            const errors_cache = reactivity_1.ref([]);
            const result = reactivity_1.computed(() => {
                errors_cache.value = errors.value;
                return cacheWithSourceMap.value;
            });
            const cacheWithSourceMap = reactivity_1.computed(() => {
                const doc = document.value;
                if (!doc)
                    return [];
                let result = toTsSourceDiags('script', errors_cache.value, doc.uri, templatetTsSourceMaps.value);
                if (onlyUnusedCheck) {
                    result = result.filter(error => { var _a; return (_a = error.tags) === null || _a === void 0 ? void 0 : _a.includes(node_1.DiagnosticTag.Unnecessary); });
                }
                return result;
            });
            return {
                result,
                cache: cacheWithSourceMap,
            };
        }
        function useTemplateScriptValidation(mode) {
            const errors_1 = reactivity_1.computed(() => {
                if (mode === 1) { // watching
                    templateTsProjectVersion.value;
                }
                const doc = templateLsTemplateScript.textDocument.value;
                if (!doc)
                    return [];
                if (mode === 1) {
                    return templateTsLs.doValidation(doc.uri, { semantic: true });
                }
                else if (mode === 2) {
                    return templateTsLs.doValidation(doc.uri, { syntactic: true });
                }
                else if (mode === 3) {
                    return templateTsLs.doValidation(doc.uri, { suggestion: true });
                }
                else if (mode === 4) {
                    return templateTsLs.doValidation(doc.uri, { declaration: true });
                }
                return [];
            });
            const errors_2 = reactivity_1.computed(() => {
                const result = [];
                if (!templateLsTemplateScript.textDocument.value
                    || !templateLsTemplateScript.teleportSourceMap.value)
                    return result;
                for (const diag of errors_1.value) {
                    const spanText = templateLsTemplateScript.textDocument.value.getText(diag.range);
                    if (!templateScriptData.setupReturns.includes(spanText))
                        continue;
                    const propRights = templateLsTemplateScript.teleportSourceMap.value.getMappedRanges(diag.range.start, diag.range.end);
                    for (const propRight of propRights) {
                        if (propRight.data.isAdditionalReference)
                            continue;
                        const definitions = templateTsLs.findDefinition(templateLsTemplateScript.textDocument.value.uri, propRight.start);
                        for (const definition of definitions) {
                            if (definition.targetUri !== templateLsScript.textDocument.value.uri)
                                continue;
                            result.push({
                                ...diag,
                                range: definition.targetSelectionRange,
                            });
                        }
                    }
                }
                return result;
            });
            const errors_1_cache = reactivity_1.ref([]);
            const errors_2_cache = reactivity_1.ref([]);
            const result = reactivity_1.computed(() => {
                errors_1_cache.value = errors_1.value;
                errors_2_cache.value = errors_2.value;
                return cacheWithSourceMap.value;
            });
            const cacheWithSourceMap = reactivity_1.computed(() => {
                const result_1 = templateLsTemplateScript.textDocument.value ? toTsSourceDiags('template', errors_1_cache.value, templateLsTemplateScript.textDocument.value.uri, templatetTsSourceMaps.value) : [];
                const result_2 = templateLsScript.textDocument.value ? toTsSourceDiags('template', errors_2_cache.value, templateLsScript.textDocument.value.uri, templatetTsSourceMaps.value) : [];
                return [...result_1, ...result_2];
            });
            return {
                result,
                cache: cacheWithSourceMap,
            };
        }
        function toSourceDiags(errors, virtualScriptUri, sourceMaps) {
            const result = [];
            for (const error of errors) {
                if (css.Diagnostic.is(error) || node_1.Diagnostic.is(error)) {
                    for (const sourceMap of sourceMaps) {
                        if (sourceMap.mappedDocument.uri !== virtualScriptUri)
                            continue;
                        const vueRange = sourceMap.getSourceRange(error.range.start, error.range.end);
                        if (!vueRange)
                            continue;
                        result.push({
                            ...error,
                            range: vueRange,
                        });
                    }
                }
            }
            return result;
        }
        function toTsSourceDiags(lsType, errors, virtualScriptUri, sourceMaps) {
            const result = [];
            for (const error of errors) {
                const vueRange = findVueRange(virtualScriptUri, error.range);
                if (vueRange) {
                    const vueError = {
                        ...error,
                        range: vueRange,
                    };
                    if (vueError.relatedInformation) {
                        const vueInfos = [];
                        for (const info of vueError.relatedInformation) {
                            const vueInfoRange = findVueRange(info.location.uri, info.location.range);
                            if (vueInfoRange) {
                                vueInfos.push({
                                    location: {
                                        uri: vueInfoRange.uri,
                                        range: vueInfoRange,
                                    },
                                    message: info.message,
                                });
                            }
                        }
                        vueError.relatedInformation = vueInfos;
                    }
                    result.push(vueError);
                }
            }
            return result;
            function findVueRange(virtualUri, virtualRange) {
                for (const sourceMap of sourceMaps) {
                    if (sourceMap.mappedDocument.uri === virtualUri) {
                        const vueRange = sourceMap.getSourceRange(virtualRange.start, virtualRange.end);
                        if (vueRange && vueRange.data.capabilities.diagnostic) {
                            return {
                                uri: vueUri,
                                start: vueRange.start,
                                end: vueRange.end,
                            };
                        }
                        // patching for ref sugar
                        // TODO: move to source map
                        const vueStartRange = sourceMap.getSourceRange(virtualRange.start);
                        if (vueStartRange && vueStartRange.data.capabilities.diagnostic) {
                            const vueEndRange = sourceMap.getSourceRange(virtualRange.end);
                            if (vueStartRange && vueEndRange && vueStartRange.data.capabilities.diagnostic && vueEndRange.data.capabilities.diagnostic) {
                                return {
                                    uri: vueUri,
                                    start: vueStartRange.start,
                                    end: vueEndRange.start,
                                };
                            }
                        }
                    }
                }
                if ('sourceFiles' in context) {
                    for (const vueLoc of context.sourceFiles.fromTsLocation(lsType, virtualUri, virtualRange.start, virtualRange.end)) {
                        if (vueLoc.type === 'source-ts' || vueLoc.range.data.capabilities.diagnostic) {
                            return {
                                uri: vueLoc.uri,
                                start: vueLoc.range.start,
                                end: vueLoc.range.end,
                            };
                        }
                    }
                }
            }
        }
    }
    function useComponentCompletionData() {
        const result = reactivity_1.computed(() => {
            { // watching
                templateTsProjectVersion.value;
            }
            const data = new Map();
            if (templateLsTemplateScript.textDocument.value && virtualTemplateRaw.textDocument.value) {
                const doc = templateLsTemplateScript.textDocument.value;
                const text = doc.getText();
                for (const tag of [...templateScriptData.componentItems, ...templateScriptData.htmlElementItems]) {
                    const tagName = tag.data.name;
                    let bind = [];
                    let on = [];
                    let slot = [];
                    {
                        const searchText = `__VLS_componentPropsBase['${tagName}']['`;
                        let offset = text.indexOf(searchText);
                        if (offset >= 0) {
                            offset += searchText.length;
                            bind = templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(offset));
                        }
                    }
                    {
                        const searchText = `__VLS_componentEmits['${tagName}']('`;
                        let offset = text.indexOf(searchText);
                        if (offset >= 0) {
                            offset += searchText.length;
                            on = templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(offset));
                        }
                    }
                    {
                        const searchText = `__VLS_components_0['${tagName}'].__VLS_slots['`;
                        let offset = text.indexOf(searchText);
                        if (offset >= 0) {
                            offset += searchText.length;
                            slot = templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(offset));
                        }
                    }
                    data.set(tagName, { item: tag, bind, on, slot });
                }
                const globalBind = templateTsLs.__internal__.doCompleteSync(doc.uri, doc.positionAt(doc.getText().indexOf(string_1.SearchTexts.GlobalAttrs)));
                data.set('*', { item: undefined, bind: globalBind, on: [], slot: [] });
            }
            return data;
        });
        return () => {
            var _a, _b;
            templateTsProjectVersion.value = (_b = (_a = templateTsLs.__internal__.host).getProjectVersion) === null || _b === void 0 ? void 0 : _b.call(_a);
            return result.value;
        };
    }
}
exports.createSourceFile = createSourceFile;
//# sourceMappingURL=sourceFile.js.map